import { TYPE } from "./constants";
import { determineType, isSerializedType } from "./common";
import type {
  SerializedError,
  SerializedRegex,
  SerializedDate,
} from "./serializers";
import {
  deserializeFunction,
  deserializeError,
  deserializePromise,
  deserializeRegex,
  deserializeDate,
  deserializeArray,
  deserializeObject,
  deserializeString,
  deserializeNumber,
  deserializeBoolean,
  deserializeNull,
  deserializeUndefined,
} from "./serializers";
import type { CustomSerializedType } from "./types";

type Deserializer<V, S> = (serializedValue: S, key: string) => V;
type PrimitiveDeserializer<V, S = V> = (serializedValue: S, key: string) => V;

type Deserializers = {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  [key: string]: Deserializer<unknown, any>;
  function: Deserializer<unknown, unknown>;
  error: Deserializer<Error, SerializedError>;
  promise: Deserializer<unknown, unknown>;
  regex: Deserializer<RegExp, SerializedRegex>;
  date: Deserializer<Date, SerializedDate>;
  array: PrimitiveDeserializer<readonly unknown[]>;
  object: PrimitiveDeserializer<Record<string, unknown>>;
  string: PrimitiveDeserializer<string>;
  number: PrimitiveDeserializer<number>;
  boolean: PrimitiveDeserializer<boolean>;
  null: PrimitiveDeserializer<undefined>;
  undefined: PrimitiveDeserializer<void>;
};

const DESERIALIZER: Deserializers = {
  [TYPE.FUNCTION]: deserializeFunction,
  [TYPE.ERROR]: deserializeError,
  [TYPE.PROMISE]: deserializePromise,
  [TYPE.REGEX]: deserializeRegex,
  [TYPE.DATE]: deserializeDate,
  [TYPE.ARRAY]: deserializeArray,
  [TYPE.OBJECT]: deserializeObject,
  [TYPE.STRING]: deserializeString,
  [TYPE.NUMBER]: deserializeNumber,
  [TYPE.BOOLEAN]: deserializeBoolean,
  [TYPE.NULL]: deserializeNull,
  [TYPE.UNDEFINED]: deserializeUndefined,
};

const defaultDeserializers: Partial<Deserializers> = {};

export function deserialize<T>(
  str: string,
  deserializers: Partial<Deserializers> = defaultDeserializers
): T {
  if (str === TYPE.UNDEFINED) {
    // the lib allows undefined returns but doesnt expect type assertions for it. need fixing
    // @ts-expect-error - we need to update the definitiona of this to allow for undefined because of this but itll take a lot of work
    return;
  }

  function replacer(
    key: string,
    val: CustomSerializedType<string, T>
  ): unknown {
    // @ts-expect-error - function this has unknown caller
    if (isSerializedType(this)) {
      return val;
    }

    let type;
    let value;

    if (isSerializedType(val)) {
      type = val.__type__;
      value = val.__val__;
    } else {
      type = determineType(val);
      value = val;
    }

    if (!type) {
      return value;
    }

    // dynamic code execution
    // eslint-disable-next-line security/detect-object-injection
    const deserializer = deserializers[type] ?? DESERIALIZER[type];

    if (!deserializer) {
      return value;
    }

    return deserializer(value, key);
  }

  // eslint-disable-next-line @typescript-eslint/no-unsafe-return
  return JSON.parse(str, replacer);
}
